@@ -137,7 +137,8 @@ request_response (GSLockPlug *plug,
         int   response;
         char *text;
 
-        gs_lock_plug_show_prompt (plug, prompt, visible);
+        gs_lock_plug_set_sensitive (plug, TRUE);
+        gs_lock_plug_enable_prompt (plug, prompt, visible);
         response = gs_lock_plug_run (plug);
 
         gs_debug ("got response: %d", response);
@@ -146,6 +147,7 @@ request_response (GSLockPlug *plug,
         if (response == GS_LOCK_PLUG_RESPONSE_OK) {
                 gs_lock_plug_get_text (plug, &text);
         }
+        gs_lock_plug_disable_prompt (plug);
 
         return text;
 }
@@ -206,16 +208,19 @@ auth_message_handler (GSAuthMessageStyle style,
 {
         gboolean    ret;
         GSLockPlug *plug;
-	const char *message;
+        const char *message;
 
         plug = GS_LOCK_PLUG (data);
 
         gs_profile_start (NULL);
         gs_debug ("Got message style %d: '%s'", style, msg);
 
+        gtk_widget_show (GTK_WIDGET (plug));
+        gs_lock_plug_set_ready (plug);
+
         ret = TRUE;
         *response = NULL;
-	message = maybe_translate_message (msg);
+        message = maybe_translate_message (msg);
 
         switch (style) {
         case GS_AUTH_MESSAGE_PROMPT_ECHO_ON:
@@ -243,10 +248,19 @@ auth_message_handler (GSAuthMessageStyle style,
         }
 
         if (*response == NULL) {
-                gs_debug ("Got not response");
+                gs_debug ("Got no response");
                 ret = FALSE;
+        } else {
+                gs_lock_plug_show_message (plug, _("Checking..."));
+                gs_lock_plug_set_sensitive (plug, FALSE);
         }
 
+        /* we may have pending events that should be processed before continuing back into PAM */
+        while (gtk_events_pending ()) {
+                gtk_main_iteration ();
+        }
+
+        gs_lock_plug_set_busy (plug);
         gs_profile_end (NULL);
 
         return ret;
@@ -255,7 +269,7 @@ auth_message_handler (GSAuthMessageStyle style,
 static gboolean
 reset_idle_cb (GSLockPlug *plug)
 {
-        gtk_widget_set_sensitive (GTK_WIDGET (plug), TRUE);
+        gs_lock_plug_set_sensitive (plug, TRUE);
         gs_lock_plug_show_message (plug, NULL);
 
         return FALSE;
@@ -269,6 +283,8 @@ do_auth_check (GSLockPlug *plug)
 
         error = NULL;
 
+        gs_lock_plug_disable_prompt (plug);
+        gs_lock_plug_set_busy (plug);
         res = gs_auth_verify_user (g_get_user_name (), g_getenv ("DISPLAY"), auth_message_handler, plug, &error);
 
         gs_debug ("Verify user returned: %s", res ? "TRUE" : "FALSE");
@@ -280,7 +296,6 @@ do_auth_check (GSLockPlug *plug)
                         gs_lock_plug_show_message (plug, _("Authentication failed."));
                 }
 
-                gtk_widget_set_sensitive (GTK_WIDGET (plug), FALSE);
                 g_timeout_add (3000, (GSourceFunc)reset_idle_cb, plug);
 
                 printf ("NOTICE=AUTH FAILED\n");
@@ -298,7 +313,8 @@ static void
 response_cb (GSLockPlug *plug,
              gint        response_id)
 {
-        if (response_id == GS_LOCK_PLUG_RESPONSE_CANCEL) {
+        if ((response_id == GS_LOCK_PLUG_RESPONSE_CANCEL) ||
+	    (response_id == GTK_RESPONSE_DELETE_EVENT)) {
                 quit_response_cancel ();
         }
 }
@@ -317,6 +333,13 @@ auth_check_idle (GSLockPlug *plug)
         return !res;
 }
 
+static void
+show_cb (GtkWidget *widget,
+         gpointer   data)
+{
+        print_id (widget);
+}
+
 static gboolean
 popup_dialog_idle (void)
 {
@@ -339,10 +362,9 @@ popup_dialog_idle (void)
         }
 
         g_signal_connect (GS_LOCK_PLUG (widget), "response", G_CALLBACK (response_cb), NULL);
+        g_signal_connect (widget, "show", G_CALLBACK (show_cb), NULL);
 
-        gtk_widget_show (widget);
-
-        print_id (widget);
+        gtk_widget_realize (widget);
 
         g_idle_add ((GSourceFunc)auth_check_idle, widget);
 
