@@ -58,7 +58,12 @@ static DBusHandlerResult gs_listener_message_handler    (DBusConnection  *connec
 #define GS_LISTENER_PATH      "/org/gnome/ScreenSaver"
 #define GS_LISTENER_INTERFACE "org.gnome.ScreenSaver"
 
-#define DBUS_INTERFACE_HAL "org.freedesktop.Hal.Device"
+#define HAL_DEVICE_INTERFACE "org.freedesktop.Hal.Device"
+
+#define CK_NAME              "org.freedesktop.ConsoleKit"
+#define CK_MANAGER_PATH      "/org/freedesktop/ConsoleKit/Manager"
+#define CK_MANAGER_INTERFACE "org.freedesktop.ConsoleKit.Manager"
+#define CK_SESSION_INTERFACE "org.freedesktop.ConsoleKit.Session"
 
 #define TYPE_MISMATCH_ERROR GS_LISTENER_INTERFACE ".TypeMismatch"
 
@@ -77,6 +82,7 @@ struct GSListenerPrivate
         GHashTable     *throttlers;
         time_t          active_start;
         time_t          session_idle_start;
+        char           *session_id;
 };
 
 typedef struct
@@ -86,6 +92,7 @@ typedef struct
         char    *reason;
         char    *connection;
         guint32  cookie;
+        GTimeVal since;
 } GSListenerRefEntry;
 
 enum {
@@ -159,13 +166,13 @@ send_dbus_message (DBusConnection *connection,
         g_return_val_if_fail (message != NULL, FALSE);
 
         if (! connection) {
-                g_warning ("There is no valid connection to the message bus");
+                gs_debug ("There is no valid connection to the message bus");
                 return FALSE;
         }
 
         is_connected = dbus_connection_get_is_connected (connection);
         if (! is_connected) {
-                g_warning ("Not connected to the message bus");
+                gs_debug ("Not connected to the message bus");
                 return FALSE;
         }
 
@@ -174,6 +181,30 @@ send_dbus_message (DBusConnection *connection,
         return sent;
 }
 
+static void
+send_dbus_boolean_signal (GSListener *listener,
+                          const char *name,
+                          gboolean    value)
+{
+        DBusMessage    *message;
+        DBusMessageIter iter;
+
+        g_return_if_fail (listener != NULL);
+
+        message = dbus_message_new_signal (GS_LISTENER_PATH,
+                                           GS_LISTENER_SERVICE,
+                                           name);
+
+        dbus_message_iter_init_append (message, &iter);
+        dbus_message_iter_append_basic (&iter, DBUS_TYPE_BOOLEAN, &value);
+
+        if (! send_dbus_message (listener->priv->connection, message)) {
+                gs_debug ("Could not send %s signal", name);
+        }
+
+        dbus_message_unref (message);
+}
+
 static void
 send_dbus_void_signal (GSListener *listener,
                        const char *name)
@@ -187,12 +218,21 @@ send_dbus_void_signal (GSListener *listener,
                                            name);
 
         if (! send_dbus_message (listener->priv->connection, message)) {
-                g_warning ("Could not send %s signal", name);
+                gs_debug ("Could not send %s signal", name);
         }
 
         dbus_message_unref (message);
 }
 
+void
+gs_listener_emit_power_notice (GSListener *listener,
+                               gboolean    in_effect)
+{
+        g_return_if_fail (listener != NULL);
+
+        send_dbus_boolean_signal (listener, "SessionPowerManagementIdleChanged", in_effect);
+}
+
 void
 gs_listener_emit_auth_request_begin (GSListener *listener)
 {
@@ -212,52 +252,20 @@ gs_listener_emit_auth_request_end (GSListener *listener)
 static void
 gs_listener_send_signal_active_changed (GSListener *listener)
 {
-        DBusMessage    *message;
-        DBusMessageIter iter;
-        dbus_bool_t     active;
-
         g_return_if_fail (listener != NULL);
 
         gs_debug ("Sending the ActiveChanged(%s) signal on the session bus",
                   listener->priv->active ? "TRUE" : "FALSE");
 
-        message = dbus_message_new_signal (GS_LISTENER_PATH,
-                                           GS_LISTENER_SERVICE,
-                                           "ActiveChanged");
-
-        active = listener->priv->active;
-        dbus_message_iter_init_append (message, &iter);
-        dbus_message_iter_append_basic (&iter, DBUS_TYPE_BOOLEAN, &active);
-
-        if (! send_dbus_message (listener->priv->connection, message)) {
-                g_warning ("Could not send ActiveChanged signal");
-        }
-
-        dbus_message_unref (message);
+        send_dbus_boolean_signal (listener, "ActiveChanged", listener->priv->active);
 }
 
 static void
 gs_listener_send_signal_session_idle_changed (GSListener *listener)
 {
-        DBusMessage    *message;
-        DBusMessageIter iter;
-        dbus_bool_t     idle;
-
         g_return_if_fail (listener != NULL);
 
-        message = dbus_message_new_signal (GS_LISTENER_PATH,
-                                           GS_LISTENER_SERVICE,
-                                           "SessionIdleChanged");
-
-        idle = listener->priv->session_idle;
-        dbus_message_iter_init_append (message, &iter);
-        dbus_message_iter_append_basic (&iter, DBUS_TYPE_BOOLEAN, &idle);
-
-        if (! send_dbus_message (listener->priv->connection, message)) {
-                g_warning ("Could not send SessionIdleChanged signal");
-        }
-
-        dbus_message_unref (message);
+        send_dbus_boolean_signal (listener, "SessionIdleChanged", listener->priv->session_idle);
 }
 
 static const char *
@@ -374,7 +382,7 @@ gs_listener_set_throttle (GSListener *listener,
         g_return_if_fail (GS_IS_LISTENER (listener));
 
         if (listener->priv->throttled != throttled) {
-                gs_debug ("Changine throttle status: %d", throttled);
+                gs_debug ("Changing throttle status: %d", throttled);
 
                 listener->priv->throttled = throttled;
 
@@ -577,7 +585,7 @@ raise_error (DBusConnection *connection,
         vsnprintf (buf, sizeof (buf), format, args);
         va_end (args);
 
-        g_warning (buf);
+        gs_debug (buf);
         reply = dbus_message_new_error (in_reply_to, error_name, buf);
         if (reply == NULL) {
                 g_error ("No memory");
@@ -601,7 +609,7 @@ raise_syntax (DBusConnection *connection,
 }
 
 static guint32
-generate_cookie ()
+generate_cookie (void)
 {
         guint32 cookie;
 
@@ -693,6 +701,91 @@ listener_remove_ref_entry (GSListener *listener,
         return removed;
 }
 
+#if GLIB_CHECK_VERSION(2,12,0)
+#define _g_time_val_to_iso8601(t) g_time_val_to_iso8601(t)
+#else
+/* copied from GLib */
+static gchar *
+_g_time_val_to_iso8601 (GTimeVal *time_)
+{
+  gchar *retval;
+
+  g_return_val_if_fail (time_->tv_usec >= 0 && time_->tv_usec < G_USEC_PER_SEC, NULL);
+
+#define ISO_8601_LEN 	21
+#define ISO_8601_FORMAT "%Y-%m-%dT%H:%M:%SZ"
+  retval = g_new0 (gchar, ISO_8601_LEN + 1);
+
+  strftime (retval, ISO_8601_LEN,
+	    ISO_8601_FORMAT,
+	    gmtime (&(time_->tv_sec)));
+
+  return retval;
+}
+#endif
+
+static void
+accumulate_ref_entry (gpointer            key,
+                      GSListenerRefEntry *entry,
+                      DBusMessageIter    *iter)
+{
+        char *description;
+        char *time;
+
+        time = _g_time_val_to_iso8601 (&entry->since);
+
+        description = g_strdup_printf ("Application=\"%s\"; Since=\"%s\"; Reason=\"%s\";",
+                                       entry->application,
+                                       time,
+                                       entry->reason);
+
+	dbus_message_iter_append_basic (iter, DBUS_TYPE_STRING, &description);
+
+        g_free (description);
+        g_free (time);
+}
+
+static DBusHandlerResult
+listener_dbus_get_ref_entries (GSListener     *listener,
+                               int             entry_type,
+                               DBusConnection *connection,
+                               DBusMessage    *message)
+{
+        DBusMessage        *reply;
+        GHashTable         *hash;
+        DBusMessageIter     iter;
+	DBusMessageIter     iter_array;
+
+        hash = get_hash_for_entry_type (listener, entry_type);
+
+        reply = dbus_message_new_method_return (message);
+        if (reply == NULL) {
+                g_error ("No memory");
+        }
+
+        dbus_message_iter_init_append (reply, &iter);
+	dbus_message_iter_open_container (&iter,
+					  DBUS_TYPE_ARRAY,
+					  DBUS_TYPE_STRING_AS_STRING,
+					  &iter_array);
+
+        if (hash != NULL) {
+                g_hash_table_foreach (hash,
+                                      (GHFunc)accumulate_ref_entry,
+                                      &iter_array);
+        }
+
+	dbus_message_iter_close_container (&iter, &iter_array);
+
+        if (! dbus_connection_send (connection, reply, NULL)) {
+                g_error ("No memory");
+        }
+
+        dbus_message_unref (reply);
+
+        return DBUS_HANDLER_RESULT_HANDLED;
+}
+
 static DBusHandlerResult
 listener_dbus_add_ref_entry (GSListener     *listener,
                              int             entry_type,
@@ -733,6 +826,7 @@ listener_dbus_add_ref_entry (GSListener     *listener,
         entry->cookie = listener_generate_unique_key (listener, entry_type);
         entry->application = g_strdup (application);
         entry->reason = g_strdup (reason);
+        g_get_current_time (&entry->since);
 
         listener_add_ref_entry (listener, entry_type, entry);
 
@@ -880,7 +974,7 @@ raise_property_type_error (DBusConnection *connection,
         snprintf (buf, 511,
                   "Type mismatch setting property with id %s",
                   device_id);
-        g_warning (buf);
+        gs_debug (buf);
 
         reply = dbus_message_new_error (in_reply_to,
                                         TYPE_MISMATCH_ERROR,
@@ -922,7 +1016,7 @@ listener_set_property (GSListener     *listener,
                         break;
                 }
         default:
-                g_warning ("Unsupported property type %d", type);
+                gs_debug ("Unsupported property type %d", type);
                 break;
         }
 
@@ -978,7 +1072,7 @@ listener_get_property (GSListener     *listener,
                 }
                 break;
         default:
-                g_warning ("Unsupported property id %u", prop_id);
+                gs_debug ("Unsupported property id %u", prop_id);
                 break;
         }
 
@@ -1013,11 +1107,11 @@ listener_get_active_time (GSListener     *listener,
 
                 if (now < listener->priv->active_start) {
                         /* shouldn't happen */
-                        g_warning ("Active start time is in the future");
+                        gs_debug ("Active start time is in the future");
                         secs = 0;
                 } else if (listener->priv->active_start <= 0) {
                         /* shouldn't happen */
-                        g_warning ("Active start time was not set");
+                        gs_debug ("Active start time was not set");
                         secs = 0;
                 } else {
                         secs = now - listener->priv->active_start;
@@ -1060,11 +1154,11 @@ listener_get_session_idle_time (GSListener     *listener,
 
                 if (now < listener->priv->session_idle_start) {
                         /* shouldn't happen */
-                        g_warning ("Session idle start time is in the future");
+                        gs_debug ("Session idle start time is in the future");
                         secs = 0;
                 } else if (listener->priv->session_idle_start <= 0) {
                         /* shouldn't happen */
-                        g_warning ("Session idle start time was not set");
+                        gs_debug ("Session idle start time was not set");
                         secs = 0;
                 } else {
                         secs = now - listener->priv->session_idle_start;
@@ -1121,6 +1215,9 @@ do_introspect (DBusConnection *connection,
                                "    <method name=\"UnInhibit\">\n"
                                "      <arg name=\"cookie\" direction=\"in\" type=\"u\"/>\n"
                                "    </method>\n"
+                               "    <method name=\"GetInhibitors\">\n"
+                               "      <arg name=\"list\" direction=\"out\" type=\"as\"/>\n"
+                               "    </method>\n"
                                "    <method name=\"Throttle\">\n"
                                "      <arg name=\"application_name\" direction=\"in\" type=\"s\"/>\n"
                                "      <arg name=\"reason\" direction=\"in\" type=\"s\"/>\n"
@@ -1150,6 +1247,9 @@ do_introspect (DBusConnection *connection,
                                "    <signal name=\"SessionIdleChanged\">\n"
                                "      <arg name=\"new_value\" type=\"b\"/>\n"
                                "    </signal>\n"
+                               "    <signal name=\"SessionPowerManagementIdleChanged\">\n"
+                               "      <arg name=\"new_value\" type=\"b\"/>\n"
+                               "    </signal>\n"
                                "    <signal name=\"AuthenticationRequestBegin\">\n"
                                "    </signal>\n"
                                "    <signal name=\"AuthenticationRequestEnd\">\n"
@@ -1217,6 +1317,9 @@ listener_dbus_handle_session_message (DBusConnection *connection,
         if (dbus_message_is_method_call (message, GS_LISTENER_SERVICE, "UnInhibit")) {
                 return listener_dbus_remove_ref_entry (listener, REF_ENTRY_TYPE_INHIBIT, connection, message);
         }
+        if (dbus_message_is_method_call (message, GS_LISTENER_SERVICE, "GetInhibitors")) {
+                return listener_dbus_get_ref_entries (listener, REF_ENTRY_TYPE_INHIBIT, connection, message);
+        }
         if (dbus_message_is_method_call (message, GS_LISTENER_SERVICE, "Throttle")) {
                 return listener_dbus_add_ref_entry (listener, REF_ENTRY_TYPE_THROTTLE, connection, message);
         }
@@ -1260,7 +1363,15 @@ listener_dbus_handle_system_message (DBusConnection *connection,
         g_return_val_if_fail (connection != NULL, DBUS_HANDLER_RESULT_NOT_YET_HANDLED);
         g_return_val_if_fail (message != NULL, DBUS_HANDLER_RESULT_NOT_YET_HANDLED);
 
-        if (dbus_message_is_signal (message, DBUS_INTERFACE_HAL, "Condition")) {
+#if 1
+        gs_debug ("obj_path=%s interface=%s method=%s destination=%s",
+                  dbus_message_get_path (message),
+                  dbus_message_get_interface (message),
+                  dbus_message_get_member (message),
+                  dbus_message_get_destination (message));
+#endif
+
+        if (dbus_message_is_signal (message, HAL_DEVICE_INTERFACE, "Condition")) {
                 DBusError   error;
                 const char *event;
                 const char *keyname;
@@ -1280,6 +1391,30 @@ listener_dbus_handle_system_message (DBusConnection *connection,
                         dbus_error_free (&error);
                 }
 
+                return DBUS_HANDLER_RESULT_HANDLED;
+        } else if (dbus_message_is_signal (message, CK_SESSION_INTERFACE, "Unlock")) {
+                const char *ssid;
+
+                ssid = dbus_message_get_path (message);
+                if (ssid != NULL
+                    && listener->priv->session_id != NULL
+                    && strcmp (ssid, listener->priv->session_id) == 0) {
+                        gs_debug ("Console kit requested session unlock");
+                        gs_listener_set_active (listener, FALSE);
+                }
+
+                return DBUS_HANDLER_RESULT_HANDLED;
+        } else if (dbus_message_is_signal (message, CK_SESSION_INTERFACE, "Lock")) {
+                const char *ssid;
+
+                ssid = dbus_message_get_path (message);
+                if (ssid != NULL
+                    && listener->priv->session_id != NULL
+                    && strcmp (ssid, listener->priv->session_id) == 0) {
+                        gs_debug ("ConsoleKit requested session lock");
+                        gs_listener_set_active (listener, TRUE);
+                }
+
                 return DBUS_HANDLER_RESULT_HANDLED;
         }
 
@@ -1339,8 +1474,8 @@ gs_listener_dbus_init (GSListener *listener)
                 listener->priv->connection = dbus_bus_get (DBUS_BUS_SESSION, &error);
                 if (listener->priv->connection == NULL) {
                         if (dbus_error_is_set (&error)) {
-                                g_warning ("couldn't connect to session bus: %s",
-                                           error.message);
+                                gs_debug ("couldn't connect to session bus: %s",
+                                          error.message);
                                 dbus_error_free (&error);
                         }
                         return FALSE;
@@ -1354,8 +1489,8 @@ gs_listener_dbus_init (GSListener *listener)
                 listener->priv->system_connection = dbus_bus_get (DBUS_BUS_SYSTEM, &error);
                 if (listener->priv->system_connection == NULL) {
                         if (dbus_error_is_set (&error)) {
-                                g_warning ("couldn't connect to system bus: %s",
-                                           error.message);
+                                gs_debug ("couldn't connect to system bus: %s",
+                                          error.message);
                                 dbus_error_free (&error);
                         }
                         return FALSE;
@@ -1687,13 +1822,70 @@ gs_listener_acquire (GSListener *listener,
 
         dbus_bus_add_match (listener->priv->system_connection,
                             "type='signal'"
-                            ",interface='"DBUS_INTERFACE_HAL"'"
+                            ",interface='"HAL_DEVICE_INTERFACE"'"
                             ",member='Condition'",
                             NULL);
+        dbus_bus_add_match (listener->priv->system_connection,
+                            "type='signal'"
+                            ",interface='"CK_SESSION_INTERFACE"'"
+                            ",member='Unlock'",
+                            NULL);
+        dbus_bus_add_match (listener->priv->system_connection,
+                            "type='signal'"
+                            ",interface='"CK_SESSION_INTERFACE"'"
+                            ",member='Lock'",
+                            NULL);
 
         return acquired;
 }
 
+static char *
+query_session_id (GSListener *listener)
+{
+        DBusMessage    *message;
+        DBusMessage    *reply;
+        DBusError       error;
+	DBusMessageIter reply_iter;
+        char           *ssid;
+
+        ssid = NULL;
+
+        dbus_error_init (&error);
+
+        message = dbus_message_new_method_call (CK_NAME, CK_MANAGER_PATH, CK_MANAGER_INTERFACE, "GetCurrentSession");
+        if (message == NULL) {
+                gs_debug ("Couldn't allocate the dbus message");
+                return NULL;
+        }
+
+        /* FIXME: use async? */
+        reply = dbus_connection_send_with_reply_and_block (listener->priv->system_connection,
+                                                           message,
+                                                           -1, &error);
+        dbus_message_unref (message);
+
+        if (dbus_error_is_set (&error)) {
+                gs_debug ("%s raised:\n %s\n\n", error.name, error.message);
+                dbus_error_free (&error);
+                return NULL;
+        }
+
+        dbus_message_iter_init (reply, &reply_iter);
+        dbus_message_iter_get_basic (&reply_iter, &ssid);
+
+        dbus_message_unref (reply);
+
+        return g_strdup (ssid);
+}
+
+static void
+init_session_id (GSListener *listener)
+{
+        g_free (listener->priv->session_id);
+        listener->priv->session_id = query_session_id (listener);
+        gs_debug ("Got session-id: %s", listener->priv->session_id);
+}
+
 static void
 gs_listener_init (GSListener *listener)
 {
@@ -1701,6 +1893,8 @@ gs_listener_init (GSListener *listener)
 
         gs_listener_dbus_init (listener);
 
+        init_session_id (listener);
+
         listener->priv->inhibitors = g_hash_table_new_full (g_int_hash,
                                                             g_int_equal,
                                                             NULL,
@@ -1731,6 +1925,8 @@ gs_listener_finalize (GObject *object)
                 g_hash_table_destroy (listener->priv->throttlers);
         }
 
+        g_free (listener->priv->session_id);
+
         G_OBJECT_CLASS (gs_listener_parent_class)->finalize (object);
 }
 
