@@ -997,7 +997,7 @@ forward_key_events (GSWindow *window)
 {
         window->priv->key_events = g_list_reverse (window->priv->key_events);
 
-        while (window->priv->key_events) {
+        while (window->priv->key_events != NULL) {
                 GdkEventKey *event = window->priv->key_events->data;
 
                 gtk_window_propagate_key_event (GTK_WINDOW (window), event);
@@ -1468,6 +1468,24 @@ gs_window_request_unlock (GSWindow *window)
         g_signal_emit (window, signals [DIALOG_UP], 0);
 }
 
+void
+gs_window_cancel_unlock_request (GSWindow  *window)
+{
+        /* FIXME: This is a bit of a hammer approach...
+	 * Maybe we should send a delete-event to
+	 * the plug?
+	 */
+        g_return_if_fail (GS_IS_WINDOW (window));
+
+	if (window->priv->lock_socket == NULL) {
+		return;
+        }
+
+	if (window->priv->lock_pid > 0) {
+		kill (window->priv->lock_pid, SIGTERM);
+        }
+}
+
 void
 gs_window_set_lock_enabled (GSWindow *window,
                             gboolean  lock_enabled)
@@ -1743,12 +1761,17 @@ static gboolean
 gs_window_real_motion_notify_event (GtkWidget      *widget,
                                     GdkEventMotion *event)
 {
-        GSWindow *window;
-        gdouble   distance;
-        gdouble   min_distance = 50;
+        GSWindow  *window;
+        gdouble    distance;
+        gdouble    min_distance;
+        gdouble    min_percentage = 0.1;
+        GdkScreen *screen;
 
         window = GS_WINDOW (widget);
 
+        screen = gs_window_get_screen (window);
+        min_distance = gdk_screen_get_width (screen) * min_percentage;
+
         /* if the last position was not set then don't detect motion */
         if (window->priv->last_x < 0 || window->priv->last_y < 0) {
                 window->priv->last_x = event->x;
